"""Format text for output in columns on the screen."""

from collections import defaultdict

def format_money(n):
    """Ledger-format dollar amounts, producing strings like '(5,138.22)'."""
    s = '%d.%02d' % divmod(abs(n), 100)
    for i in range(-6, -len(s), -3):
        s = s[:i] + ',' + s[i:] # put in thousands commas
    if n < 0:
        return '(%s)' % s
    else:
        return '%s ' % s # trailing space keeps numbers aligned

class Table(object):
    """Accept a list of rows, and print them with aligned columns."""

    def __init__(self):
        self.lines = []
        self.fieldwidths = defaultdict(int)

    def append(self, args):
        """Add a new row of fields, given as zero or more arguments."""
        # We also update our record of how wide each column needs to be.
        fields = []
        fieldwidths = self.fieldwidths
        for i, field in enumerate(args):
            if isinstance(field, int):
                field = format_money(field)
            elif field is None:
                field = ''
            fieldwidths[i] = max(fieldwidths[i], len(field))
            fields.append(field)
        self.lines.append(fields)

    def display(self):
        """Print out all of the rows of data we were given."""
        fieldwidths = self.fieldwidths
        lastfield = len(fieldwidths) - 1
        for line in self.lines:
            for i, field in enumerate(line):
                if field == '-':
                    print ' ' + '-' * fieldwidths[i],
                elif i < lastfield:
                    print ' %*s' % (fieldwidths[i], field or ''),
                else:
                    print ' ' + (field or ''),
            print

def display(rows):
    """Given a sequence of rows, print them as a table."""
    t = Table()
    for row in rows:
        t.append(row)
    t.display()



"""The command-line interface to running cashflow."""

import sys
from optparse import OptionParser

from cashflow import open as cashflow_open
from cashflow.format import display
from cashflow.reports import cashflow

def main():
    """Command-line cashflow program (an entry point in `setup.py`)."""

    # Check the command line for validity.

    parser = OptionParser(usage="usage: %prog `filename`")
    (options, args) = parser.parse_args()

    if len(args) != 1:
        parser.error("you must provide the `filename` of a Gnucash file")

    # Open our input file.

    try:
        book = cashflow_open(args[0])
    except IOError, e:
        sys.stderr.write("cashflow: cannot open file %r: %s\n"
                         % (args[0], e.strerror or e.message))
        sys.exit(1)

    # Run the report and print it out.

    rows = cashflow(book)
    display(rows)



"""Models that represent GnuCash data.

These classes are lightweight wrappers around the actual ElementTree
that holds the data from inside a GnuCash save file.  Typically the user
simply needs to parse the file and instantiate a Book() around its tree
object; a convenience function for doing this is defined at the top of
this package.

"""
from cashflow.oxm import XMLWrapper, XMLChild, XMLChildText, XMLChildren

# Precompute the names of common GnuCash XML tags.

def tag(colontag):
    """Return the fully-qualified XML tag name for a GnuCash tag."""
    prefix, name = colontag.split(':')
    return '{http://www.gnucash.org/XML/%s}%s' % (prefix, name)

gnc_book = tag('gnc:book')
gnc_account = tag('gnc:account')
gnc_transaction = tag('gnc:transaction')
act_id = tag('act:id')
act_name = tag('act:name')
act_type = tag('act:type')
trn_date_posted = tag('trn:date-posted')
trn_splits = tag('trn:splits')
trn_split = tag('trn:split')
ts_date = tag('ts:date')
split_value = tag('split:value')
split_account = tag('split:account')

# The actual classes used to expose a GnuCash file.

class Split(XMLWrapper):
    """A GnuCash split (one of several parts of a transaction)."""
    value = XMLChildText(split_value)
    account_guid = XMLChildText(split_account)

class Splits(XMLWrapper):
    """A GnuCash splits element."""
    splits = XMLChildren(trn_split, Split)

    def __iter__(self):
        return self.splits

class Transaction(XMLWrapper):
    """A GnuCash transaction."""
    date_posted = XMLChildText(trn_date_posted, ts_date)
    splits = XMLChild(trn_splits, Splits)

class Account(XMLWrapper):
    """A GnuCash account."""
    guid = XMLChildText(act_id)
    name = XMLChildText(act_name)
    type = XMLChildText(act_type)

class Book(object):
    """A GnuCash ledger book (the contents of a GnuCash save file)."""

    def __init__(self, tree):
        # init method saves the tree object, to support write() someday
        self.tree = tree
        self.element = tree.find(tag('gnc:book'))

    accounts = XMLChildren(gnc_account, Account)
    transactions = XMLChildren(gnc_transaction, Transaction)



"""Cash flow analysis for GnuCash accounts."""

import gzip
import xml.etree.cElementTree as ET
from cashflow.book import Book

def open(path):
    """Open a GnuCash file and return a Book object that manages it."""
    tree = ET.parse(gzip.open(path))
    return Book(tree)

def sample_file_path():
    """Return the path to the Gnucash file included as a sample."""
    import os.path
    return os.path.join(os.path.dirname(__file__), 'sample.gnucash')



"""Analyze 
"""

class Results(object):
    """Set of results for where your cash went every month."""

    

class Analyzer(object):
    """Class that runs through your accounts looking at your cash flow."""

    def __init__(self):
        self.account_map = {}

    def map(self, from_account, to_account):
        """Rewrite ``from_account`` to ``to_account``."""
        self.account_map[from_account] = to_account

    def analyze(self, book):
        pass



"""Various reports that can be generated from a GnuCash ledger book."""

from collections import defaultdict

def make_int_dict():
    """A dictionary whose default values are (integer) zero."""
    return defaultdict(int)

def sorted_rows(rows):
    """Return the sequence of rows, sorted by the 4th column."""
    def fourth_tuple_item(t):
        return t[3]
    return sorted(rows, key=fourth_tuple_item)

def cashflow(book, one_month=None, cash_types=('BANK', 'CASH', 'CREDIT'),
             account_map={}, is_cash=[], isnt_cash=[], regular_expenses=[]):
    """Print a cash flow report for the given month, or all months."""
    month_regular = defaultdict(make_int_dict)
    month_expenses = defaultdict(make_int_dict)
    month_income = defaultdict(make_int_dict)

    # Fetch the accounts from inside this GnuCash ledger book.

    account_list = [ a for a in book.accounts ]

    # First, determine by what name each account should be known, and
    # create a map from account guids (which are what we will find named
    # in transaction splits) to printable account names; we include any
    # transforms specified in our account_map parameter.  This might
    # result in several accounts being mapped to the same name.

    account_names = {}
    for account in account_list:
        name = account.name
        while name in account_map:
            name = account_map[name]
        account_names[account.guid] = name

    # Next, we need to determine which accounts count as our cash pool.

    accounts = dict( (a.guid, a) for a in account_list )
    cash_account_guids = set( a.guid for a in account_list
                              if a.type in cash_types)
    for name in is_cash:
        aa = [ a for a in account_list if a.name == name ]
        if not aa:
            raise ValueError('no account has the name %r' % name)
        if len(aa) > 1:
            raise ValueError('%d accounts have the name %r' % (len(aa), name))
        cash_account_guids.add(aa[0].guid)

    for name in isnt_cash:
        aa = [ a for a in account_list if a.name == name ]
        if not aa:
            raise ValueError('no account has the name %r' % name)
        if len(aa) > 1:
            raise ValueError('%d accounts have the name %r' % (len(aa), name))
        #print '---------', cash_account_guids
        cash_account_guids.remove(aa[0].guid)
        #print '---------', cash_account_guids

    for t in book.transactions:
        #print t.date_posted
        month = t.date_posted[:7] # string like '2008-08'
        #print month
        splits = list(t.splits)
        if any( (s.account_guid in cash_account_guids) for s in splits ):
            for split in splits:
                guid = split.account_guid
                if guid not in cash_account_guids:
                    name = account_names[guid]
                    value = split.value
                    ivalue = int(value[:value.find('/')])
                    if ivalue > 0:
                        if name in regular_expenses:
                            month_regular[month][name] -= ivalue
                        else:
                            month_expenses[month][name] -= ivalue
                    elif ivalue < 0:
                        month_income[month][name] -= ivalue

    rows = []

    months = set()
    months.update(month_expenses)
    months.update(month_income)
    months = sorted(months)

    t = 0
    for month in months:
        if one_month is None or month == one_month:
            t2 = 0

            if month_income[month]:
                rows.append(())
                t3 = 0
                for name, total in sorted(month_income[month].items()):
                    rows.append((None, None, total, name))
                    t3 += total
                rows.append((None, None, '-'))
                rows.append((None, t3, t3, 'Income for ' + month))
                t2 += t3

            if month_regular[month]:
                rows.append(())
                t3 = 0
                for name, total in sorted(month_regular[month].items()):
                    rows.append((None, None, total, name))
                    t3 += total
                rows.append((None, None, '-'))
                rows.append((None, t3, t3, 'Regular expenses for ' + month))
                t2 += t3

            if month_expenses[month]:
                rows.append(())
                t3 = 0
                for name, total in sorted(month_expenses[month].items()):
                    rows.append((None, None, total, name))
                    t3 += total
                rows.append((None, None, '-'))
                rows.append((None, t3, t3, 'Expenses for ' + month))
                t2 += t3

            rows.append((None, '-'))
            rows.append((t2, t2, '', 'Monthly total for ' + month))
            t += t2
            rows.append(('-'))
            rows.append((t, '', '', 'Running total after ' + month))

    return rows



from setuptools import setup

import os

def read(*filenames):
    return open(os.path.join(os.path.dirname(__file__), *filenames)).read()

setup(name = 'cashflow',
      version = '1.0',
      description = 'Reads a Gnucash file and computes your cash flow',
      long_description = read('src/cashflow/README.txt'),
      license = 'GPL',
      author = 'Brandon Craig Rhodes',
      author_email = 'brandon@rhodesmill.org',
      classifiers = [
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Financial and Insurance Industry',
        'Environment :: Console',
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Topic :: Office/Business',
        'Topic :: Office/Business :: Financial',
        'Topic :: Office/Business :: Financial :: Accounting',
        ],
      packages = [ 'cashflow' ],
      package_dir = { '': 'src' },
      test_suite = 'cashflow.tests',
      entry_points = {
        'console_scripts': [
            'cashflow = cashflow.command:main',
            ],
        },
      )



"""Object-XML mapper.

These base classes and descriptors can be used to wrap ElementTree nodes
and make them behave like normal Python objects.  Think of this as being
like an ORM but for XML data.

"""
class XMLWrapper(object):
    """A class that wraps an XML object to expose its sub-elements."""

    def __init__(self, element):
        self.element = element

class XMLChild(object):
    """Descriptor returning a single XML child element."""

    def __init__(self, tag, cls):
        self.tag = tag
        self.cls = cls

    def __get__(self, instance, owner):
        return self.cls(instance.element.find(self.tag))

class XMLChildText(object):
    """Descriptor for a text attribute in a child XML element."""

    def __init__(self, *tags):
        self.tags = tags

    def __get__(self, instance, owner):
        e = instance.element
        for tag in self.tags:
            e = e.find(tag)
        return e.text

class XMLChildren(object):
    """Descriptor returning an iterator over XML child elements."""

    def __init__(self, tag, cls):
        self.tag = tag
        self.cls = cls

    def __get__(self, instance, owner):
        cls = self.cls
        return ( cls(e) for e in instance.element.findall(self.tag) )
